# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.import proto
"""Defines queries that are accessible via `builtin` namespace.

Built-in queries provides quick access to important elements in Google Ads API
with a simple SQL syntax.
All built-in queries are available in `builtin` namespace
(i.e. `builtin.ocid_mapping`) and can be added to SQL scripts
(i.e. `SELECT * FROM builtin.ocid_mapping`).

"""
from __future__ import annotations

import functools
import operator
import re
from collections import abc
from collections import defaultdict
from typing import Callable

import gaarf
from gaarf import query_executor


def get_ocid_mapping(report_fetcher: query_executor.AdsReportFetcher,
                     accounts: list[str]) -> gaarf.report.GaarfReport:
    """Returns mapping between external customer_id and OCID parameter.


    OCID parameter is used to build links to Google Ads entities in UI.

    Args:
        report_fetcher: An instance of AdsReportFetcher.
        accounts: Google Ads accounts to get data on OCID mapping.

    Returns:
        Report with mapping between external customer_id and OCID parameter.
    """
    query = ('SELECT customer.id AS account_id, '
             'metrics.optimization_score_url  AS url FROM customer')
    mapping = []
    for account in accounts:
        if (report := report_fetcher.fetch(query, account)):
            for row in report:
                if (ocid := re.findall(r'ocid=(\w+)', row.url)):
                    mapping.append([row.account_id, ocid[0]])
                    break
            if not ocid:
                mapping.append([int(account), '0'])
        else:
            mapping.append([int(account), '0'])
    return gaarf.report.GaarfReport(
        results=mapping, column_names=[
            'account_id',
            'ocid',
        ])


def get_account_hierarchy_flattened(
        report_fetcher: query_executor.AdsReportFetcher,
        accounts: list[str]) -> gaarf.report.GaarfReport:
    SEED_ACCOUNTS_QUERY = """
    SELECT
        customer_client.level AS level,
        customer_client.manager AS is_manager,
        customer_client.id AS account_id
    FROM customer_client
    WHERE customer_client.level <= 1
    """

    NESTED_ACCOUNTS_QUERY = """
    SELECT
        customer.id AS mcc_id,
        customer.descriptive_name AS mcc_name,
        {level} AS level,
        customer_client_link.client_customer~0 AS account_id
    FROM customer_client_link
    """

    seed_accounts = accounts if isinstance(accounts,
                                           abc.MutableSequence) else [accounts]
    hierarchy = report_fetcher.fetch(SEED_ACCOUNTS_QUERY, seed_accounts)
    level_mapping = defaultdict(list)
    for row in hierarchy:
        if row.is_manager:
            level_mapping[row.level].append(row.account_id)

    reports: list[gaarf.report.GaarfReport] = []
    for level, accounts in level_mapping.items():
        if accounts:
            report_fetcher.customer_ids = accounts
            reports.append(
                report_fetcher.fetch(NESTED_ACCOUNTS_QUERY.format(level=level)))

    combined_report = functools.reduce(operator.add, reports)
    df = combined_report.to_pandas()
    max_level = max(df.level)
    root_mcc_df = df[df.level == 0][['mcc_id', 'mcc_name', 'account_id']]
    root_mcc_df = root_mcc_df.rename(columns={
        'mcc_id': 'root_mcc_id',
        'mcc_name': 'root_mcc_name'
    })
    root_mcc_df['account_id'] = root_mcc_df['account_id'].astype(int)
    selectable_columns = ['root_mcc_id', 'root_mcc_name']
    if max_level == 0:
        return gaarf.report.GaarfReport.from_pandas(root_mcc_df)
    else:
        for i in range(1, max(df.level) + 1):
            if i == 1:
                n_mcc_df = df[df.level == i][[
                    'mcc_id', 'mcc_name', 'account_id'
                ]]
                mcc_ids = n_mcc_df.merge(
                    root_mcc_df,
                    how='right',
                    left_on='mcc_id',
                    right_on='account_id')[[
                        'root_mcc_id', 'root_mcc_name', 'account_id_x',
                        'mcc_id', 'mcc_name', 'account_id_y'
                    ]]
            else:
                previous_level = i - 1
                n_minus_1_mcc_df = df[df.level == previous_level][[
                    'mcc_id', 'mcc_name', 'account_id'
                ]]
                n_minus_1_mcc_df = n_minus_1_mcc_df.rename(
                    columns={
                        'mcc_id': f'level_{previous_level}_mcc_id',
                        'mcc_name': f'level_{previous_level}_mcc_name',
                    })
                mcc_ids = n_mcc_df.merge(
                    n_minus_1_mcc_df,
                    how='right',
                    left_on='mcc_id',
                    right_on='account_id')[[
                        f'level_{previous_level}_mcc_id',
                        f'level_{previous_level}_mcc_name', 'account_id_x',
                        'mcc_id', 'mcc_name', 'account_id_y'
                    ]]

            mcc_ids.mcc_name.fillna('Direct', inplace=True)
            mcc_ids.account_id_x.fillna(mcc_ids.account_id_y, inplace=True)
            mcc_ids.mcc_id.fillna(mcc_ids.account_id_y, inplace=True)
            mcc_ids = mcc_ids.rename(
                columns={
                    'mcc_name': f'level_{i}_mcc_name',
                    'account_id_x': 'account_id',
                    'mcc_id': f'level_{i}_mcc_id'
                })[selectable_columns +
                   [f'level_{i}_mcc_id', f'level_{i}_mcc_name', 'account_id']]
            mcc_ids[f'level_{i}_mcc_id'] = mcc_ids[f'level_{i}_mcc_id'].astype(
                int)
        mcc_ids['account_id'] = mcc_ids['account_id'].astype(int)
        return gaarf.report.GaarfReport.from_pandas(mcc_ids)


BUILTIN_QUERIES: dict[str, Callable] = {
    'ocid_mapping': get_ocid_mapping,
    'account_hierarchy_flattened': get_account_hierarchy_flattened
}



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Module for executing Gaarf queries and writing them to local/remote.


Module defines two major classes:
    * AdsReportFetcher - to perform fetching data from Ads API, parsing it
      and returning GaarfReport.
    * AdsQueryExecutor - to perform fetching data from Ads API in a form of
      GaarfReport and saving it to local/remote storage.
"""
from __future__ import annotations

import enum
import importlib
import itertools
import logging
import warnings
from collections.abc import MutableSequence
from collections.abc import Sequence
from concurrent import futures
from typing import Any
from typing import Generator

from gaarf import api_clients
from gaarf import builtin_queries
from gaarf import exceptions
from gaarf import parsers
from gaarf import query_editor
from gaarf import report
from gaarf.io.writers import abs_writer
from gaarf.io.writers import console_writer
from google.ads.googleads import errors as googleads_exceptions
from google.api_core import exceptions as google_exceptions

google_ads_service = importlib.import_module(
    f'google.ads.googleads.{api_clients.GOOGLE_ADS_API_VERSION}.'
    'services.types.google_ads_service')

logger = logging.getLogger(__name__)


class OptimizeStrategy(enum.Enum):
    """Specifies how to parse response from Google Ads API."""
    NONE = 1
    BATCH = 2
    PROTOBUF = 3
    BATCH_PROTOBUF = 4


class AdsReportFetcher:
    """Class responsible for getting data from Ads API.

    Attributes:
        api_client: a client used for connecting to Ads API.
    """

    def __init__(self,
                 api_client: api_clients.BaseClient,
                 customer_ids: Sequence[str] | None = None) -> None:
        self.api_client = api_client
        if customer_ids:
            warnings.warn(
                '`AdsReportFetcher` will deprecate passing `customer_ids` to `__init__` method. '
                'Consider passing list of customer_ids to `AdsReportFetcher.fetch` method',
                category=DeprecationWarning,
                stacklevel=3)
            self.customer_ids = [
                customer_ids
            ] if not isinstance(customer_ids, list) else customer_ids

    def expand_mcc(self,
                   customer_ids: str | MutableSequence,
                   customer_ids_query: str | None = None) -> list[str]:
        return self._get_customer_ids(customer_ids, customer_ids_query)

    def fetch(self,
              query_specification: str | query_editor.QueryElements,
              customer_ids: list[str] | str | None = None,
              customer_ids_query: str | None = None,
              expand_mcc: bool = False,
              args: dict[str, Any] | None = None,
              optimize_strategy: str = 'NONE') -> report.GaarfReport:
        """Fetches data from Ads API based on query_specification.

        Args:
            query_specification: Query text that will be passed to Ads API
                alongside column_names, customizers and virtual columns.
            customer_ids: Account(s) for from which data should be fetched.
            custom_query: GAQL query used to reduce the number of customer_ids.
            expand_mcc: Whether to perform expansion of root customer_ids
                into leaf accounts.
            args: Arguments that need to be passed to the query
            optimize_strategy: strategy for speeding up query execution
                ("NONE", "PROTOBUF", "BATCH", "BATCH_PROTOBUF").

        Returns:
            GaarfReport with results of query execution.

        Raises:
            GaarfExecutorException:
                When customer_ids are not provided or Ads API returned error.
            GaarfBuiltInQueryException:
                When built-in query cannot be found in the registry.
        """

        if isinstance(self.api_client, api_clients.GoogleAdsApiClient):
            if not customer_ids:
                warnings.warn(
                    '`AdsReportFetcher` will require passing `customer_ids` '
                    'to `fetch` method.',
                    category=DeprecationWarning,
                    stacklevel=3)
                if hasattr(self, 'customer_ids'):
                    if not self.customer_ids:
                        raise exceptions.GaarfExecutorException(
                            'Please specify add `customer_ids` to '
                            '`fetch` method')
                    customer_ids = self.customer_ids
            else:
                if expand_mcc:
                    customer_ids = self.expand_mcc(customer_ids,
                                                   customer_ids_query)
                customer_ids = [
                    customer_ids
                ] if not isinstance(customer_ids, list) else customer_ids
        else:
            customer_ids = []
        total_results: list[list[tuple]] = []
        if not isinstance(query_specification, query_editor.QueryElements):
            query_specification = query_editor.QuerySpecification(
                text=str(query_specification),
                args=args,
                api_version=self.api_client.api_version).generate()

        if query_specification.is_builtin_query:
            if not (builtin_report := builtin_queries.BUILTIN_QUERIES.get(
                    query_specification.query_title)):
                raise exceptions.GaarfBuiltInQueryException(
                    'Cannot find the built-in query '
                    f'"{query_specification.title}"')
            return builtin_report(self, accounts=customer_ids)
        optimize_strategy = OptimizeStrategy[optimize_strategy]
        parser = parsers.GoogleAdsRowParser(query_specification)
        for customer_id in customer_ids:
            logger.debug('Running query %s for customer_id %s',
                         query_specification.query_title, customer_id)
            try:
                results = self._parse_ads_response(query_specification,
                                                   customer_id, parser,
                                                   optimize_strategy)
                total_results.extend(results)
                if query_specification.is_constant_resource:
                    logger.debug('Constant resource query: running only once')
                    break
            except googleads_exceptions.GoogleAdsException as e:
                logger.error('Cannot execute query %s for %s',
                             query_specification.query_title, customer_id)
                logger.error(str(e))
                raise exceptions.GaarfExecutorException
        if not total_results:
            results_placeholder = [
                parser.parse_ads_row(self.api_client.google_ads_row)
            ]
            if not isinstance(self.api_client, api_clients.BaseClient):
                logger.warning(
                    'Query %s generated zero results, '
                    'using placeholders to infer schema',
                    query_specification.query_title)
        else:
            results_placeholder = []
        return report.GaarfReport(
            results=total_results,
            column_names=query_specification.column_names,
            results_placeholder=results_placeholder,
            query_specification=query_specification)

    def _parse_ads_response(
        self,
        query_specification: query_editor.QueryElements,
        customer_id: str,
        parser: parsers.GoogleAdsRowParser,
        optimize_strategy: OptimizeStrategy = OptimizeStrategy.NONE
    ) -> list[list[tuple]]:
        """Parses response returned from Ads API request.

        Args:
            query_specification:
                Query text that will be passed to Ads API
                alongside column_names, customizers and virtual columns.
            customer_id:
                Account for which data should be requested.
            parser:
                An instance of parser class that transforms each row from
                request into desired format.
            optimize_strategy:
                Strategy for speeding up query execution
                ("NONE", "PROTOBUF", "BATCH", "BATCH_PROTOBUF").

        Returns:
            Parsed rows for the whole response.

        Raises:
            google_exceptions.InternalServerError:
                When data cannot be fetched from Ads API.
        """
        logger.debug('Getting response for query %s for customer_id %s',
                     query_specification.query_title, customer_id)
        try:
            response = self.api_client.get_response(
                entity_id=str(customer_id),
                query_text=query_specification.query_text,
                query_title=query_specification.query_title)
        except google_exceptions.InternalServerError:
            logging.error('Cannot fetch data from API for query "%s" 3 times',
                          query_specification.query_title)
            raise

        if optimize_strategy in (OptimizeStrategy.BATCH,
                                 OptimizeStrategy.BATCH_PROTOBUF):
            logger.warning('Running gaarf in an optimized mode')
            logger.warning('Optimize strategy is %s', optimize_strategy.name)
        if optimize_strategy in (OptimizeStrategy.BATCH,
                                 OptimizeStrategy.BATCH_PROTOBUF):
            return self._parse_ads_response_in_batches(
                response,
                query_specification,
                customer_id,
                parser,
            )
        return self._parse_ads_response_sequentially(response,
                                                     query_specification,
                                                     customer_id, parser)

    def _parse_ads_response_in_batches(
            self, response: google_ads_service.SearchGoogleAdsResponse,
            query_specification: query_editor.QueryElements, customer_id: str,
            parser: parsers.GoogleAdsRowParser) -> list[list]:
        """Parses response returned from Ads API request in parallel batches.

        Args:
            response: Google Ads API response.
            query_specification:
                Query text that will be passed to Ads API
                alongside column_names, customizers and virtual columns.
            customer_id:
                Account for which data are parsed.
            parser:
                An instance of parser class that transforms each row from
                request into desired format.

        Returns:
            Parsed rows for the whole response.
        """
        parsed_batches = []
        with futures.ThreadPoolExecutor() as executor:
            future_to_batch = {}
            for batch in response:
                future_to_batch[executor.submit(self._parse_batch, parser,
                                                batch.results)] = batch.results
            for i, future in enumerate(
                    futures.as_completed(future_to_batch), start=1):
                logger.debug('Parsed batch %d for query %s for customer_id %s',
                             i, query_specification.query_title, customer_id)
                parsed_batch = future.result()
                parsed_batches.append(parsed_batch)
        return list(itertools.chain.from_iterable(parsed_batches))

    def _parse_ads_response_sequentially(
            self, response: google_ads_service.SearchGoogleAdsResponse,
            query_specification: query_editor.QueryElements, customer_id: str,
            parser: parsers.GoogleAdsRowParser) -> list[list]:
        """Parses response returned from Ads API request sequentially.

        Args:
            response: Google Ads API response.
            query_specification:
                Query text that will be passed to Ads API
                alongside column_names, customizers and virtual columns.
            customer_id:
                Account for which data are parsed.
            parser:
                An instance of parser class that transforms each row from
                request into desired format.

        Returns:
            Parsed rows for the whole response.
        """
        total_results: list[list] = []
        logger.debug('Iterating over response for query %s for customer_id %s',
                     query_specification.query_title, customer_id)
        for batch in response:
            logger.debug('Parsing batch for query %s for customer_id %s',
                         query_specification.query_title, customer_id)

            results = self._parse_batch(parser, batch.results)
            total_results.extend(list(results))
        return total_results

    def _parse_batch(
        self, parser: parsers.GoogleAdsRowParser,
        batch: Sequence[google_ads_service.GoogleAdsRow]
    ) -> Generator[list, None, None]:
        """Parse reach row from batch of Ads API response.

        Args:
            parser:
                An instance of parser class that transforms each row from
                request into desired format.
            batch:
                Sequence of GoogleAdsRow that needs to be parsed.
        Yields:
            Parsed rows for a batch.
        """

        for row in batch:
            yield parser.parse_ads_row(row)

    def _get_customer_ids(self,
                          seed_customer_ids: str | MutableSequence,
                          customer_ids_query: str | None = None) -> list[str]:
        """Gets list of customer_ids from an MCC account.

        Args:
            customer_ids: MCC account_id(s).
            custom_query: GAQL query used to reduce the number of customer_ids.
        Returns:
            All customer_ids from MCC satisfying the condition.
        """

        query = """
        SELECT customer_client.id FROM customer_client
        WHERE customer_client.manager = FALSE
        AND customer_client.status = ENABLED
        """
        query_specification = query_editor.QuerySpecification(query).generate()
        if not isinstance(seed_customer_ids, MutableSequence):
            seed_customer_ids = seed_customer_ids.split(',')
        child_customer_ids = self.fetch(query_specification,
                                        seed_customer_ids).to_list()
        if customer_ids_query:
            query_specification = query_editor.QuerySpecification(
                customer_ids_query).generate()
            child_customer_ids = self.fetch(query_specification,
                                            child_customer_ids)
            child_customer_ids = [
                row[0] if isinstance(row, report.GaarfRow) else row
                for row in child_customer_ids
            ]

        child_customer_ids = list(
            set([
                customer_id for customer_id in child_customer_ids
                if customer_id != 0
            ]))

        return child_customer_ids


class AdsQueryExecutor:
    """Class responsible for getting data from Ads API and writing it to
        local/remote storage.

    Attributes:
        api_client: a client used for connecting to Ads API.
    """

    def __init__(self, api_client: api_clients.BaseClient) -> None:
        """Initializes QueryExecutor.

        Args:
            api_client: a client used for connecting to Ads API.
        """
        self.api_client = api_client

    @property
    def report_fetcher(self) -> AdsReportFetcher:
        """Initializes AdsReportFetcher to get data from Ads API."""
        return AdsReportFetcher(self.api_client)

    def execute(self,
                query_text: str,
                query_name: str,
                customer_ids: list[str] | str,
                writer_client: abs_writer.AbsWriter = console_writer
                .ConsoleWriter(),
                args: dict[str, Any] | None = None,
                optimize_performance: str = 'NONE') -> None:
        """Reads query, extract results and stores them in a specified location.

        Args:
            query_text: Text for the query.
            customer_ids: All accounts for which query will be executed.
            writer_client: Client responsible for writing data to local/remote
                location.
            args: Arguments that need to be passed to the query
            optimize_strategy: strategy for speeding up query execution
                ("NONE", "PROTOBUF", "BATCH", "BATCH_PROTOBUF")
        """

        query_specification = query_editor.QuerySpecification(
            query_text, query_name, args,
            self.report_fetcher.api_client.api_version).generate()
        results = self.report_fetcher.fetch(
            query_specification=query_specification,
            customer_ids=customer_ids,
            optimize_strategy=optimize_performance)
        logger.debug('Start writing data for query %s via %s writer',
                     query_specification.query_title, type(writer_client))
        writer_client.write(results, query_specification.query_title)
        logger.debug('Finish writing data for query %s via %s writer',
                     query_specification.query_title, type(writer_client))

    def expand_mcc(self,
                   customer_ids: str | MutableSequence,
                   customer_ids_query: str | None = None) -> list[str]:
        """Performs Manager account(s) expansion to child accounts.

        Args:
            customer_ids:
                Manager account(s) to be expanded.
            customer_ids_query:
                Gaarf query to limit the expansion only to accounts
                satisfying the condition.

        Returns:
            All child accounts under provided customer_ids.
        """
        return self.report_fetcher._get_customer_ids(customer_ids,
                                                     customer_ids_query)



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

import pytest
from gaarf.io import writer


@pytest.mark.parametrize('option', ['bq', 'bigquery'])
def test_create_writer_returns_correct_fields_for_bq_option(option):
    bq_writer = writer.create_writer(
        option, project='fake_project', dataset='fake_dataset')
    assert bq_writer.dataset_id == 'fake_project.fake_dataset'


def test_create_writer_returns_correct_fields_for_csv_option():
    csv_writer = writer.create_writer('csv', destination_folder='/fake_folder')
    assert csv_writer.destination_folder == '/fake_folder'


def test_create_writer_returns_correct_fields_for_sheets_option():
    sheet_writer = writer.create_writer(
        'sheet',
        share_with='1@google.com',
        credentials_file='home/me/client_secret.json')
    assert sheet_writer.share_with == '1@google.com'
    assert sheet_writer.credentials_file == 'home/me/client_secret.json'


def test_create_writer_returns_correct_fields_for_sqldb_option():
    sqlalchemy_writer = writer.create_writer(
        'sqldb', connection_string='protocol://user:password@host:port/db')
    assert sqlalchemy_writer.connection_string == (
        'protocol://user:password@host:port/db')


def test_create_writer_returns_correct_fields_for_json_option():
    json_writer = writer.create_writer(
        'json', destination_folder='/fake_folder')
    assert json_writer.destination_folder == '/fake_folder'


def test_null_writer_raises_unknown_writer_error():
    with pytest.raises(ValueError):
        writer.create_writer('non-existing-option')


class TestDeprecatedWriters:

    def test_instantiating_writer_factory_is_deprected(self):
        with pytest.deprecated_call():
            writer.WriterFactory()

    def test_instantiating_console_writer_is_deprected(self):
        with pytest.warns(DeprecationWarning) as w:
            writer.StdoutWriter()
            assert len(w) == 1
            assert str(w[0].message) == (
                'Loading StdoutWriter from `gaarf.io.writer` is deprecated; '
                'Import ConsoleWriter from `gaarf.io.writers.console_writer` '
                'instead')

    def test_instantiating_csv_writer_is_deprected(self):
        with pytest.warns(DeprecationWarning) as w:
            writer.CsvWriter()
            assert len(w) == 1
            assert issubclass(w[0].category, DeprecationWarning)
            assert str(w[0].message) == (
                'Loading CsvWriter from `gaarf.io.writer` is deprecated; '
                'Import CsvWriter from `gaarf.io.writers.csv_writer` instead')

    def test_instantiating_sqlalchemy_writer_is_deprected(self):
        with pytest.warns(DeprecationWarning) as w:
            writer.SqlAlchemyWriter(connection_string='fake-connection')
            assert len(w) == 1
            assert issubclass(w[0].category, DeprecationWarning)
            assert str(w[0].message) == (
                'Loading SqlAlchemyWriter from `gaarf.io.writer` is '
                'deprecated; '
                'Import SqlAlchemyWriter from `gaarf.io.writers.sqldb_writer` '
                'instead')

    def test_instantiating_sheet_writer_is_deprected(self):
        with pytest.warns(DeprecationWarning) as w:
            writer.SheetWriter(share_with='fake-user', credentials_file=None)
            assert len(w) == 1
            assert issubclass(w[0].category, DeprecationWarning)
            assert str(w[0].message) == (
                'Loading SheetWriter from `gaarf.io.writer` is deprecated; '
                'Import SheetWriter from `gaarf.io.writers.sheets_writer` '
                'instead')

    def test_instantiating_bigquery_writer_is_deprected(self):
        with pytest.warns(DeprecationWarning) as w:
            writer.BigQueryWriter(
                project='fake-project', dataset='fake-dataset')
            assert len(w) == 1
            assert issubclass(w[0].category, DeprecationWarning)
            assert str(w[0].message) == (
                'Loading BigQueryWriter from `gaarf.io.writer` is deprecated; '
                'Import BigQueryWriter from `gaarf.io.writers.bigquery_writer` '
                'instead')



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Module for various helpers for executing Gaarf as CLI tool."""
from __future__ import annotations

import argparse
import dataclasses
import datetime
import logging
import os
import sys
import traceback
from collections.abc import MutableSequence
from collections.abc import Sequence
from typing import Any
from typing import Callable

import rich
import smart_open
import yaml
from dateutil import relativedelta
from gaarf import query_editor
from gaarf.io import writer
from google.ads.googleads import errors as google_ads_errors


@dataclasses.dataclass
class BaseConfig:
    """Base config to inherit other configs from."""


@dataclasses.dataclass
class GaarfConfig(BaseConfig):
    """Stores values to run gaarf from command line.

    Attributes:
        output:
            Specifies where to store fetched data (console, csv, BQ.)
        api_version:
            Google Ads API version.
        account:
            Account(s) to get data from.
        params:
            Any parameters passed to Gaarf query for substitution.
        writer_params:
            Any parameters that can be passed to writer for data saving.
        customer_ids_query:
            Query text to limit accounts fetched from Ads API.
        customer_ids_query_file:
            Path to query to limit accounts fetched from Ads API.
    """
    output: str
    api_version: str
    account: str | list[str] | None
    params: dict[str, Any] = dataclasses.field(default_factory=dict)
    writer_params: dict[str, Any] = dataclasses.field(default_factory=dict)
    customer_ids_query: str | None = None
    customer_ids_query_file: str | None = None

    def __post_init__(self) -> None:
        """Ensures that values passed during __init__ correctly formatted."""
        if isinstance(self.account, MutableSequence):
            self.account = [
                account.replace('-', '').strip() for account in self.account
            ]
        else:
            self.account = self.account.replace(
                '-', '').strip() if self.account else None
        self.writer_params = {
            key.replace('-', '_'): value
            for key, value in self.writer_params.items()
        }


class GaarfConfigException(Exception):
    """Exception for invalid GaarfConfig."""


@dataclasses.dataclass
class GaarfBqConfig(BaseConfig):
    """Stores values to run gaarf-bq from command line.

    Attributes:
        project:
           Google Cloud project name.
        dataset_location:
            Location of BigQuery dataset.
        params:
            Any parameters passed to BigQuery query for substitution.
    """
    project: str
    dataset_location: str | None
    params: dict[str, Any] = dataclasses.field(default_factory=dict)


class GaarfBqConfigException(Exception):
    """Exception for invalid GaarfBqConfig."""


@dataclasses.dataclass
class GaarfSqlConfig(BaseConfig):
    """Stores values to run gaarf-sql from command line.

    Attributes:
        connection_string:
           Connection string to SqlAlchemy database engine.
        params:
            Any parameters passed to SQL query for substitution.
    """
    connection_string: str
    params: dict[str, Any] = dataclasses.field(default_factory=dict)


class GaarfSqlConfigException(Exception):
    """Exception for invalid GaarfSqlConfig."""


class BaseConfigBuilder:
    """Base class for building CLI-based Gaarf configs.

    Attributes:
        args: Parameters to initialize config.
        type: Type of config (either 'file' or 'cli').
        gaarf_config_path: Path to config file.
    """

    def __init__(self, args: tuple[argparse.Namespace, list]) -> None:
        """Initializes config based on provided arguments.

        Args:
            args: Parameters to initialize config.
        """
        self.args = args
        if (gaarf_config_path := self.args[0].gaarf_config):
            self.type = 'file'
            self.gaarf_config_path = gaarf_config_path
        else:
            self.type = 'cli'
            self.gaarf_config_path = None

    def build(self) -> BaseConfig:
        """Loads config from file or build from arguments.

        Returns:
            A subclass of BaseConfig.
        """
        if self.type == 'file':
            return self._load_gaarf_config()
        return self._build_gaarf_config()

    def _load_gaarf_config(self) -> BaseConfig:
        """Loads config from gaarf_config_path."""
        raise NotImplementedError

    def _build_gaarf_config(self) -> BaseConfig:
        """Builds config from instance args."""
        raise NotImplementedError


class GaarfConfigBuilder(BaseConfigBuilder):
    """Builds GaarfConfig from CLI.

    Attributes:
        args: Parameters to initialize config.
        type: Type of config (either 'file' or 'cli').
        gaarf_config_path: Path to config file.
    """

    def _load_gaarf_config(self) -> GaarfConfig:
        """Loads config from gaarf_config_path.

        Returns:
            GaarfConfig with values taken from config file.

        Raises:
            GaarfConfigException:
                If config file missing `gaarf` section.
            ValueError:
                If config file missing `output` section.
        """
        with smart_open.open(self.gaarf_config_path, encoding='utf-8') as f:
            config = yaml.safe_load(f)
        gaarf_section = config.get('gaarf')
        if not gaarf_section:
            raise GaarfConfigException(
                'Invalid config, must have `gaarf` section!')
        if not (output := gaarf_section.get('output')):
            raise ValueError('Config does not contains `output` section!')
        return GaarfConfig(
            output=gaarf_section.get('output'),
            api_version=gaarf_section.get('api_version'),
            account=gaarf_section.get('account'),
            params=gaarf_section.get('params', {}),
            writer_params=gaarf_section.get(output),
            customer_ids_query=gaarf_section.get('customer_ids_query'),
            customer_ids_query_file=gaarf_section.get(
                'customer_ids_query_file'))

    def _build_gaarf_config(self) -> GaarfConfig:
        """Builds config from args parameter.

        Returns:
            GaarfConfig with values taken from instance args.
        """
        main_args, query_args = self.args[0], self.args[1]
        params = ParamsParser([
            'macro',
            'template',
            main_args.save,
        ]).parse(query_args)
        return GaarfConfig(
            output=main_args.save,
            api_version=main_args.api_version,
            account=main_args.customer_id,
            params=params,
            writer_params=params.get(main_args.save),
            customer_ids_query=main_args.customer_ids_query,
            customer_ids_query_file=main_args.customer_ids_query_file)


class GaarfBqConfigBuilder(BaseConfigBuilder):
    """Builds GaarfBqConfig from CLI.

    Attributes:
        args: Parameters to initialize config.
        type: Type of config (either 'file' or 'cli').
        gaarf_config_path: Path to config file.
    """

    def _load_gaarf_config(self) -> GaarfBqConfig:
        """Loads config from gaarf_config_path.

        Returns:
            GaarfBqConfig with values taken from config file.

        Raises:
            GaarfBqConfigException:
                If config file missing `gaarf-bq` section.
        """
        with smart_open.open(self.gaarf_config_path, encoding='utf-8') as f:
            config = yaml.safe_load(f)
        gaarf_section = config.get('gaarf-bq')
        if not gaarf_section:
            raise GaarfBqConfigException(
                'Invalid config, must have `gaarf-bq` section!')
        params = gaarf_section.get('params', {})
        return GaarfBqConfig(
            project=gaarf_section.get('project'),
            dataset_location=gaarf_section.get('dataset_location'),
            params=params)

    def _build_gaarf_config(self) -> GaarfBqConfig:
        """Builds config from args parameter.

        Returns:
            GaarfBqConfig with values taken from instance args.
        """
        main_args, query_args = self.args[0], self.args[1]
        params = ParamsParser([
            'macro',
            'sql',
            'template',
        ]).parse(query_args)
        return GaarfBqConfig(
            project=main_args.project,
            dataset_location=main_args.dataset_location,
            params=params)


class GaarfSqlConfigBuilder(BaseConfigBuilder):
    """Builds GaarfSqlConfig from CLI.

    Attributes:
        args: Parameters to initialize config.
        type: Type of config (either 'file' or 'cli').
        gaarf_config_path: Path to config file.
    """

    def _load_gaarf_config(self) -> BaseConfig:
        """Loads config from gaarf_config_path.

        Returns:
            GaarfBqConfig with values taken from config file.

        Raises:
            GaarfSqlConfigException:
                If config file missing `gaarf-sql` section.
        """
        with smart_open.open(self.gaarf_config_path, encoding='utf-8') as f:
            config = yaml.safe_load(f)
        gaarf_section = config.get('gaarf-sql')
        if not gaarf_section:
            raise GaarfSqlConfigException(
                'Invalid config, must have `gaarf-sql` section!')
        params = gaarf_section.get('params', {})
        return GaarfSqlConfig(
            connection_string=gaarf_section.get('connection-string').format(
                **dict(os.environ.items())),
            params=params)

    def _build_gaarf_config(self) -> GaarfSqlConfig:
        """Builds config from args parameter.

        Returns:
            GaarfSqlConfig with values taken from instance args.
        """
        main_args, query_args = self.args[0], self.args[1]
        params = ParamsParser(['macro', 'sql', 'template']).parse(query_args)
        return GaarfSqlConfig(
            connection_string=main_args.connection_string.format(
                **dict(os.environ.items())),
            params=params)


class ParamsParser(query_editor.CommonParametersMixin):

    def __init__(self, identifiers: Sequence[str]) -> None:
        self.identifiers = identifiers

    def parse(self, params: Sequence) -> dict[str, dict | None]:
        return {
            identifier: self._parse_params(identifier, params)
            for identifier in self.identifiers
        }

    def _parse_params(self, identifier: str, params: Sequence[Any]) -> dict:
        parsed_params = {}
        if params:
            raw_params = [param.split('=', maxsplit=1) for param in params]
            for param in raw_params:
                param_pair = self._identify_param_pair(identifier, param)
                if param_pair:
                    parsed_params.update(param_pair)
        return parsed_params

    def _identify_param_pair(self, identifier: str,
                             param: Sequence[str]) -> dict[str, Any] | None:
        key = param[0]
        if identifier not in key:
            return None
        provided_identifier, key = key.split('.')
        if provided_identifier.replace('--', '') not in self.identifiers:
            raise GaarfParamsException(
                f'CLI argument {provided_identifier} is not supported'
                f", supported arguments {', '.join(self.identifiers)}")
        key = key.replace('-', '_')
        if len(param) == 2:
            # TODO: b/337860595 - Ensure that writer params are converted to int
            return {key: param[1]}
        raise GaarfParamsException(
            f'{identifier} {key} is invalid,'
            f'--{identifier}.key=value is the correct format')


class GaarfParamsException(Exception):
    """Defines exception for incorrect parameters."""


def convert_date(date_string: str) -> str:
    """Converts specific dates parameters to actual dates.

    Returns:
        Date string in YYYY-MM-DD format.

    Raises:
        ValueError:
            If dynamic lookback value (:YYYYMMDD-N) is incorrect.
    """

    if isinstance(date_string, list) or date_string.find(':YYYY') == -1:
        return date_string
    current_date = datetime.date.today()
    date_object = date_string.split('-')
    base_date = date_object[0]
    if len(date_object) == 2:
        try:
            days_ago = int(date_object[1])
        except ValueError as e:
            raise ValueError(
                'Must provide numeric value for a number lookback period, '
                'i.e. :YYYYMMDD-1') from e
    else:
        days_ago = 0
    if base_date == ':YYYY':
        new_date = datetime.datetime(current_date.year, 1, 1)
        delta = relativedelta.relativedelta(years=days_ago)
    elif base_date == ':YYYYMM':
        new_date = datetime.datetime(current_date.year, current_date.month, 1)
        delta = relativedelta.relativedelta(months=days_ago)
    elif base_date == ':YYYYMMDD':
        new_date = current_date
        delta = relativedelta.relativedelta(days=days_ago)
    return (new_date - delta).strftime('%Y-%m-%d')


class ConfigSaver:

    def __init__(self, path: str) -> None:
        self.path = path

    def save(self, gaarf_config: BaseConfig):
        if os.path.exists(self.path):
            with smart_open.open(self.path, 'r', encoding='utf-8') as f:
                config = yaml.safe_load(f)
        else:
            config = {}
        config = self.prepare_config(config, gaarf_config)
        with smart_open.open(self.path, 'w', encoding='utf-8') as f:
            yaml.dump(
                config,
                f,
                default_flow_style=False,
                sort_keys=False,
                encoding='utf-8')

    def prepare_config(self, config: dict, gaarf_config: BaseConfig) -> dict:
        gaarf = dataclasses.asdict(gaarf_config)
        if isinstance(gaarf_config, GaarfConfig):
            gaarf[gaarf_config.output] = gaarf_config.writer_params
            if not isinstance(gaarf_config.account, MutableSequence):
                gaarf['account'] = gaarf_config.account.split(',')
            del gaarf['writer_params']
            if gaarf_config.writer_params:
                del gaarf['params'][gaarf_config.output]
            gaarf = _remove_empty_values(gaarf)
            config.update({'gaarf': gaarf})
        if isinstance(gaarf_config, GaarfBqConfig):
            gaarf = _remove_empty_values(gaarf)
            config.update({'gaarf-bq': gaarf})
        if isinstance(gaarf_config, GaarfSqlConfig):
            gaarf = _remove_empty_values(gaarf)
            config.update({'gaarf-sql': gaarf})
        return config


def initialize_runtime_parameters(config: BaseConfig) -> BaseConfig:
    for key, param in config.params.items():
        for key_param, value_param in param.items():
            config.params[key][key_param] = convert_date(value_param)
        config.params[key].update(ParamsParser.common_params)
    return config


def _remove_empty_values(dict_object: dict[str, Any]) -> dict[str, Any]:
    """Remove all empty elements: strings, dictionaries from a dictionary."""
    if isinstance(dict_object, dict):
        return {
            key: value
            for key, value in ((key, _remove_empty_values(value))
                               for key, value in dict_object.items())
            if value
        }
    if isinstance(dict_object, (int, str, MutableSequence)):
        return dict_object


def gaarf_runner(query: str, callback: Callable, logger) -> None:
    try:
        logger.debug('starting query %s', query)
        callback()
        logger.info('%s executed successfully', query)
    except writer.ZeroRowException:
        logger.warning('%s returns 0 rows', query)
    except google_ads_errors.GoogleAdsException as ex:
        logger.error(
            '%s failed with status %s and includes the following errors:',
            query,
            ex.error.code().name)
        for error in ex.failure.errors:
            logger.error('\tError with message %s .', error.message)
            if error.location:
                for field in error.location.field_path_elements:
                    logger.error('\t\tOn field %s', field.field_name)
    except Exception as e:
        traceback.print_tb(e.__traceback__)
        logger.error('%s generated an exception: %s', query, str(e))


def postprocessor_runner(query: str, callback: Callable, logger) -> None:
    try:
        logger.debug('starting query %s', query)
        callback()
        logger.info('%s executed successfully', query)
    except Exception as e:
        logger.error('%s generated an exception: %s', query, str(e))


def init_logging(loglevel: str = 'INFO',
                 logger_type: str = 'local',
                 name: str = __name__) -> logging.Logger:
    if logger_type == 'rich':
        logging.basicConfig(
            format='%(message)s',
            level=loglevel,
            datefmt='%Y-%m-%d %H:%M:%S',
            handlers=[
                rich.logging.RichHandler(rich_tracebacks=True),
            ])
    else:
        logging.basicConfig(
            format='[%(asctime)s][%(name)s][%(levelname)s] %(message)s',
            stream=sys.stdout,
            level=loglevel,
            datefmt='%Y-%m-%d %H:%M:%S')
    logging.getLogger('google.ads.googleads.client').setLevel(logging.WARNING)
    logging.getLogger('smart_open.smart_open_lib').setLevel(logging.WARNING)
    logging.getLogger('urllib3.connectionpool').setLevel(logging.WARNING)
    return logging.getLogger(name)



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
from concurrent import futures
import functools

from gaarf.io import reader  # type: ignore
from gaarf.bq_executor import BigQueryExecutor
from .utils import (GaarfBqConfigBuilder, ConfigSaver,
                    initialize_runtime_parameters, postprocessor_runner,
                    init_logging)


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("query", nargs="+")
    parser.add_argument("-c", "--config", dest="gaarf_config", default=None)
    parser.add_argument("--project", dest="project")
    parser.add_argument("--dataset-location",
                        dest="dataset_location",
                        default=None)
    parser.add_argument("--save-config",
                        dest="save_config",
                        action="store_true")
    parser.add_argument("--no-save-config",
                        dest="save_config",
                        action="store_false")
    parser.add_argument("--config-destination",
                        dest="save_config_dest",
                        default="config.yaml")
    parser.add_argument("--log", "--loglevel", dest="loglevel", default="info")
    parser.add_argument("--logger", dest="logger", default="local")
    parser.add_argument("--dry-run", dest="dry_run", action="store_true")
    parser.add_argument("--parallel-queries",
                        dest="parallel_queries",
                        action="store_true")
    parser.add_argument("--no-parallel-queries",
                        dest="parallel_queries",
                        action="store_false")
    parser.set_defaults(save_config=False)
    parser.set_defaults(dry_run=False)
    parser.set_defaults(parallel_queries=True)
    args = parser.parse_known_args()
    main_args = args[0]

    logger = init_logging(loglevel=main_args.loglevel.upper(),
                          logger_type=main_args.logger)
    config = GaarfBqConfigBuilder(args).build()
    logger.debug("config: %s", config)
    if main_args.save_config and not main_args.gaarf_config:
        ConfigSaver(main_args.save_config_dest).save(config)
    if main_args.dry_run:
        exit()

    config = initialize_runtime_parameters(config)
    logger.debug("initialized config: %s", config)

    bq_executor = BigQueryExecutor(project_id=config.project,
                                   location=config.dataset_location)
    bq_executor.create_datasets(config.params.get("macro"))

    reader_client = reader.FileReader()

    if main_args.parallel_queries:
        logger.info("Running queries in parallel")
        with futures.ThreadPoolExecutor() as executor:
            future_to_query = {
                executor.submit(bq_executor.execute, query,
                                reader_client.read(query), config.params): query
                for query in sorted(main_args.query)
            }
            for future in futures.as_completed(future_to_query):
                query = future_to_query[future]
                postprocessor_runner(query, future.result, logger)
    else:
        logger.info("Running queries sequentially")
        for query in sorted(main_args.query):
            callback = functools.partial(bq_executor.execute, query,
                reader_client.read(query), config.params)
            postprocessor_runner(query, callback, logger)


if __name__ == "__main__":
    main()



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Module for installing gaarf as a package."""
from __future__ import annotations

import itertools
import pathlib

import setuptools

HERE = pathlib.Path(__file__).parent

README = (HERE / 'README.md').read_text()

EXTRAS_REQUIRE = {
    'sqlalchemy': ['sqlalchemy',],
    'simulator': ['Faker',],
    'sheets': ['gspread',],
    'bq': [
        'google-cloud-bigquery',
        'pyarrow>=14.0.1',
    ]
}
EXTRAS_REQUIRE['full'] = list(set(itertools.chain(*EXTRAS_REQUIRE.values())))

setuptools.setup(
    name='google-ads-api-report-fetcher',
    version='1.13.4',
    python_requires='>3.8',
    description=('Library for fetching reports from Google Ads API '
                 'and saving them locally & remotely.'),
    long_description=README,
    long_description_content_type='text/markdown',
    url='https://github.com/google/ads-api-reports-fetcher',
    author='Google Inc. (gTech gPS CSE team)',
    author_email='no-reply@google.com',
    license='Apache 2.0',
    classifiers=[
        'Programming Language :: Python :: 3',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Operating System :: OS Independent',
        'License :: OSI Approved :: Apache Software License',
    ],
    packages=setuptools.find_packages(),
    install_requires=[
        'google-ads>=24.0.0',
        'smart_open[all]',
        'jinja2',
        'python-dateutil',
        'pandas>=1.3.4',
        'rich',
        'tenacity',
    ],
    extras_require=EXTRAS_REQUIRE,
    setup_requires=[
        'pytest-runner',
    ],
    tests_requires=[
        'pytest',
        'pytest-mock',
    ],
    entry_points={
        'console_scripts': [
            'gaarf=gaarf.cli.gaarf:main',
            'gaarf-py=gaarf.cli.gaarf:main',
            'gaarf-bq=gaarf.cli.bq:main',
            'gaarf-py-bq=gaarf.cli.bq:main',
            'gaarf-simulator=gaarf.cli.simulator:main',
            'gaarf-sql=gaarf.cli.sql:main',
        ]
    })



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

import dataclasses
import itertools
import logging

import pytest
import tenacity
from gaarf import api_clients
from gaarf import parsers
from gaarf import query_editor
from gaarf import query_executor
from google.api_core import exceptions as google_exceptions

_QUERY = 'SELECT customer.id AS customer_id FROM customer'
_EXPECTED_RESULTS = [
    [1],
    [2],
    [3],
]


@dataclasses.dataclass
class FakeResponse:
    data: list[list[parsers.GoogleAdsRowElement]]

    def __iter__(self):
        for result in self.data:
            yield FakeBatch(result)


@dataclasses.dataclass
class FakeBatch:
    results: list[list]


@dataclasses.dataclass
class Customer:
    id: int


@dataclasses.dataclass
class FakeGoogleAdsRowElement:
    customer: Customer


class TestAdsReportFetcher:

    @pytest.fixture
    def fake_response(self):
        fake_results = [
            [
                FakeGoogleAdsRowElement(Customer(1)),
            ],
            [
                FakeGoogleAdsRowElement(Customer(2)),
            ],
            [
                FakeGoogleAdsRowElement(Customer(3)),
            ],
        ]
        return FakeResponse(data=fake_results)

    @pytest.fixture
    def test_client(self, mocker):
        mocker.patch('google.ads.googleads.client.oauth2', return_value=[])
        return api_clients.GoogleAdsApiClient()

    @pytest.fixture
    def failing_api_client(self, mocker):
        mocker.patch('google.ads.googleads.client.oauth2', return_value=[])
        mocker.patch(
            f'google.ads.googleads.{api_clients.GOOGLE_ADS_API_VERSION}'
            '.services.services.google_ads_service.GoogleAdsServiceClient'
            '.search_stream',
            side_effect=[
                google_exceptions.InternalServerError('test'),
                google_exceptions.InternalServerError('test'),
                google_exceptions.InternalServerError('test'),
            ])
        api_client = api_clients.GoogleAdsApiClient()
        api_client.get_response.retry.wait = tenacity.wait_none()
        return api_client

    @pytest.fixture
    def fake_report_fetcher(self, mocker, test_client, fake_response):
        mocker.patch(
            'gaarf.api_clients.GoogleAdsApiClient.get_response',
            return_value=fake_response)
        return query_executor.AdsReportFetcher(test_client)

    def test_parse_ads_response_returns_success(self, fake_report_fetcher,
                                                caplog):
        caplog.set_level(logging.DEBUG)
        query_specification = query_editor.QuerySpecification(_QUERY).generate()
        parser = parsers.GoogleAdsRowParser(query_specification)
        customer_id = 1
        results = fake_report_fetcher._parse_ads_response(
            query_specification=query_specification,
            parser=parser,
            customer_id=customer_id)

        assert results == _EXPECTED_RESULTS
        assert 'Getting response for query None for customer_id 1' in (
            caplog.text)

    @pytest.mark.parametrize('strategy', [
        'BATCH',
        'BATCH_PROTOBUF',
    ])
    def test_parse_ads_response_in_batches_generates_warning_messages(
            self, fake_report_fetcher, strategy, caplog):
        query_specification = query_editor.QuerySpecification(_QUERY).generate()
        parser = parsers.GoogleAdsRowParser(query_specification)
        customer_id = 1
        optimize_strategy = query_executor.OptimizeStrategy[strategy]
        fake_report_fetcher._parse_ads_response(
            query_specification=query_specification,
            parser=parser,
            customer_id=customer_id,
            optimize_strategy=optimize_strategy)

        assert 'Running gaarf in an optimized mode' in caplog.text
        assert f'Optimize strategy is {optimize_strategy.name}' in caplog.text

    def test_parse_ads_response_sequentially_returns_success(
            self, fake_report_fetcher, fake_response):
        query_specification = query_editor.QuerySpecification(_QUERY).generate()
        parser = parsers.GoogleAdsRowParser(query_specification)
        customer_id = 1
        results = fake_report_fetcher._parse_ads_response_sequentially(
            response=fake_response,
            query_specification=query_specification,
            parser=parser,
            customer_id=customer_id)

        assert results == _EXPECTED_RESULTS

    def test_parse_ads_response_in_batches_returns_success(
            self, fake_report_fetcher, fake_response):
        query_specification = query_editor.QuerySpecification(_QUERY).generate()
        parser = parsers.GoogleAdsRowParser(query_specification)
        customer_id = 1
        results = fake_report_fetcher._parse_ads_response_in_batches(
            response=fake_response,
            query_specification=query_specification,
            parser=parser,
            customer_id=customer_id)

        assert_sequence_content_the_same(results, _EXPECTED_RESULTS)

    def test_parse_ads_response_raises_internal_server_error_after_3_failed_attemps(  # pylint: disable=line-too-long
            self, failing_api_client, caplog):
        report_fetcher = query_executor.AdsReportFetcher(failing_api_client)
        query_specification = query_editor.QuerySpecification(_QUERY).generate()
        parser = parsers.GoogleAdsRowParser(query_specification)
        customer_id = 1
        with pytest.raises(google_exceptions.InternalServerError):
            report_fetcher._parse_ads_response(
                query_specification=query_specification,
                parser=parser,
                customer_id=customer_id)
        assert ('Cannot fetch data from API for query '
                f'"{query_specification.query_title}" 3 times') in caplog.text


def assert_sequence_content_the_same(
        results: list[list[parsers.GoogleAdsRowElement]],
        other_results: list[list[parsers.GoogleAdsRowElement]]) -> bool:
    return sorted(itertools.chain.from_iterable(results)) == sorted(
        itertools.chain.from_iterable(other_results))



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Module for formatting Gaarf reports before writing."""
from __future__ import annotations

from enum import Enum
from pathlib import Path
from typing import Callable
from typing import get_args
from typing import Union

import proto  # type: ignore
from gaarf.report import GaarfReport
from typing_extensions import TypeAlias

_NESTED_FIELD: TypeAlias = Union[
    list,
    proto.marshal.collections.repeated.RepeatedComposite,
    proto.marshal.collections.repeated.Repeated,
]


class FormattingStrategy:
    """Interface for all formatting strategies applied to GaarfReport."""

    def apply_transformations(self, report: GaarfReport) -> GaarfReport:
        """Applies class transformation to report."""
        raise NotImplementedError

    def _cast_to_enum(self, enum: type[Enum], value: str | Enum) -> Enum:
        """Ensures that strings are always converted to Enums."""
        return enum[value.upper()] if isinstance(value, str) else value


class ArrayHandling(Enum):
    """Specifies acceptable options for ArrayHandlingStrategy."""
    STRINGS = 1
    ARRAYS = 2


class ArrayHandlingStrategy(FormattingStrategy):
    """Handles arrays in the report.

    Arrays can be left as-is or converted to strings with required delimiter.

    Attributes:
        type_: Type of array handling (ARRAYS, STRINGS).
        delimiter: Symbol used as delimiter when type_ is STRINGS.
    """

    def __init__(self,
                 type_: ArrayHandling | str = ArrayHandling.STRINGS,
                 delimiter: str = '|') -> None:
        """Initializes strategy based on type_ and delimiter.

        Args:
            type_: Type of array handling (ARRAYS, STRINGS).
            delimiter: Symbol used as delimiter when type_ is STRINGS.
        """
        self.type_ = self._cast_to_enum(ArrayHandling, type_)
        self.delimiter = delimiter

    def apply_transformations(self, report: GaarfReport) -> GaarfReport:
        """Replaces arrays in the report."""
        if self.type_ == ArrayHandling.ARRAYS:
            return report

        formatted_rows = self._format_rows(report.results, self._delimiter_join)
        formatted_placeholders = self._format_rows(report.results_placeholder,
                                                   lambda x: '')
        return GaarfReport(
            results=formatted_rows,
            column_names=report.column_names,
            results_placeholder=formatted_placeholders)

    def _format_rows(self, rows: list[list],
                     nested_field_handler: Callable) -> list[list]:
        """Formats rows of report based on join_strategy.

        Args:
            rows: Rows on GaarfReport.
            nested_field_handler: Handlers to nested structures.

        Returns:
            Formatted rows.
        """
        formatted_rows = []
        for row in rows:
            formatted_row = []
            for field in row:
                if isinstance(field, get_args(_NESTED_FIELD)):
                    field = nested_field_handler(field)
                formatted_row.append(field)
            formatted_rows.append(formatted_row)
        return formatted_rows

    def _delimiter_join(self, field: _NESTED_FIELD) -> str:
        """Helper function to perform join by an instance delimiter.

        Args:
            field: A nested field.

        Returns:
            The same field but concatenated to string.
        """
        return self.delimiter.join([str(element) for element in field])


def format_report_for_writing(
        report: GaarfReport,
        formatting_strategies: list[FormattingStrategy]) -> GaarfReport:
    """Applies formatting strategies to report.

    Args:
        report: Report that needs to be formatted.
        formatting_strategies: Strategies to be applied to report.

    Returns:
        New report with updated data.
    """
    for strategy in formatting_strategies:
        report = strategy.apply_transformations(report)
    return report


def format_extension(path_object: str,
                     current_extension: str = '.sql',
                     new_extension: str = '') -> str:
    """Formats query path to required extension.

    Args:
        path_object: Path to query.
        current_extension: Extension of the query.
        new_extension: Required extension

    Returns:
       Path with an updated extension.
    """
    path_object_name = Path(path_object).name
    if len(path_object_name.split('.')) > 1:
        return path_object_name.replace(current_extension, new_extension)
    return f'{path_object}{new_extension}'



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Module defining key query elements for building and parsing GAQL query."""
from __future__ import annotations

import ast
import dataclasses
import datetime
import operator
import re

from dateutil import relativedelta
from gaarf import api_clients
from gaarf import exceptions
from gaarf import query_post_processor

VALID_VIRTUAL_COLUMN_OPERATORS = (
    ast.BinOp,
    ast.UnaryOp,
    ast.operator,
    ast.Num,
    ast.Expression,
)


@dataclasses.dataclass(frozen=True)
class VirtualColumn:
    type: str
    value: str
    fields: list[str] | None = None
    substitute_expression: str | None = None


@dataclasses.dataclass
class ExtractedLineElements:
    fields: list[str] | None
    alias: str | None
    virtual_column: VirtualColumn | None


@dataclasses.dataclass
class ProcessedField:
    field: str
    customizer_type: str | None = None
    customizer_value: str | None = None


@dataclasses.dataclass
class QueryElements:
    """Contains raw query and parsed elements.

    Attributes:
        query_title: Title of the query that needs to be parsed.
        query_text: Text of the query that needs to be parsed.
        resource_name: Name of Google Ads API reporting resource.
        fields: Ads API fields that need to be feched.
        column_names: Friendly names for fields which are used when saving data
        customizers: Attributes of fields that need to be be extracted.
        virtual_columns: Attributes of fields that need to be be calculated.
        is_constant_resource: Whether resource considered a constant one.
        is_builtin_query: Whether query is built-in.
    """
    query_title: str
    query_text: str
    resource_name: str
    fields: list[str] | None = None
    column_names: list[str] | None = None
    customizers: dict[str, dict[str, str]] | None = None
    virtual_columns: dict[str, VirtualColumn] | None = None
    is_constant_resource: bool = False
    is_builtin_query: bool = False


class CommonParametersMixin:
    common_params = {
        'date_iso':
            datetime.date.today().strftime('%Y%m%d'),
        'yesterday_iso':
            (datetime.date.today() -
             relativedelta.relativedelta(days=1)).strftime('%Y%m%d'),
        'current_date':
            datetime.date.today().strftime('%Y-%m-%d'),
        'current_datetime':
            datetime.datetime.today().strftime('%Y-%m-%d %H:%M:%S')
    }


class QuerySpecification(CommonParametersMixin,
                         query_post_processor.PostProcessorMixin):

    def __init__(self,
                 text: str,
                 title: str | None = None,
                 args: dict | None = None,
                 api_version: str = api_clients.GOOGLE_ADS_API_VERSION) -> None:
        self.text = text
        self.title = title
        self.args = args or {}
        self.macros = self._init_macros()
        self.base_client = api_clients.BaseClient(api_version)

    def _init_macros(self) -> dict[str, str]:
        if not self.args:
            return self.common_params
        if macros := self.args.get('macro'):
            macros.update(self.common_params)
            return macros
        return self.common_params

    def generate(self) -> QueryElements:
        """Generates necessary query elements based on query text and arguments.

        Returns:
            Various elements parsed from a query (text, fields,
            column_names, etc).
        """

        query_text = self.expand_jinja(self.text, self.args.get('template'))
        resource_name = self.extract_resource_from_query(query_text)
        if is_builtin_query := bool(resource_name.startswith('builtin')):
            return QueryElements(
                query_title=resource_name.replace('builtin.', ''),
                query_text=query_text,
                fields=None,
                column_names=None,
                customizers=None,
                virtual_columns=None,
                resource_name=resource_name,
                is_constant_resource=False,
                is_builtin_query=True)
        if not is_builtin_query and resource_name not in dir(
                self.base_client.google_ads_row):
            raise exceptions.GaarfResourceException(
                f'Invalid resource specified in the query: {resource_name}')
        is_constant_resource = bool(resource_name.endswith('_constant'))
        query_lines = self.cleanup_query_text(query_text)
        query_text = ' '.join(query_lines)
        try:
            query_text = query_text.format(**self.macros)
        except KeyError as e:
            raise exceptions.GaarfMacroException(
                f'No value provided for macro {str(e)}.') from e
        fields = []
        column_names = []
        customizers = {}
        virtual_columns = {}

        query_lines = self.extract_query_lines(' '.join(query_lines))
        for line in query_lines:
            if not line:
                continue
            field_name = None
            customizer_type = None
            line_elements = self._extract_line_elements(line, self.macros)
            if field_elements := line_elements.fields:
                for field_element in field_elements:
                    processed_field = self._process_field(field_element)
                    field_name = processed_field.field.strip().replace(',', '')
                    fields.append(self.format_type_field_name(field_name))
                    customizer_type = processed_field.customizer_type
            if not line_elements.alias and not field_name and not is_builtin_query:
                raise exceptions.GaarfVirtualColumnException(
                    'Virtual attributes should be aliased')
            column_name = line_elements.alias.strip().replace(
                ',', '') if line_elements.alias else field_name
            if column_name:
                column_names.append(self.normalize_column_name(column_name))

            if virtual_column := line_elements.virtual_column:
                virtual_columns[column_name] = virtual_column
            if customizer_type:
                customizers[column_name] = {
                    'type': customizer_type,
                    'value': processed_field.customizer_value
                }
        query_text = self.create_query_text(fields, virtual_columns, query_text)
        for _, column in virtual_columns.items():
            if not isinstance(column.value, (int, float)):
                column.value.format(**self.macros)
        return QueryElements(
            query_title=self.title,
            query_text=query_text,
            fields=fields,
            column_names=column_names,
            customizers=customizers,
            virtual_columns=virtual_columns,
            resource_name=resource_name,
            is_constant_resource=is_constant_resource,
            is_builtin_query=is_builtin_query)

    def create_query_text(self, fields: list[str],
                          virtual_columns: dict[str, VirtualColumn],
                          query_text: str) -> str:
        """Generate valid GAQL query.

        Based on original Gaarf query text, a set of field and virtual columns
        constructs new GAQL query to be sent to Ads API.

        Args:
            fields:
                All fields that need to be fetched from API.
            virtual_columns:
                Virtual columns that might contain extra fields for fetching.
            query_text: Original Gaarf query text.
        Returns:
            Description of return.
        """
        virtual_fields = [
            field for name, column in virtual_columns.items()
            if column.type == 'expression' for field in column.fields
        ]
        if virtual_fields:
            fields = fields + virtual_fields
        query_text = (f"SELECT {', '.join(fields)} "
                      f'{self.extract_from_statement(query_text)}')
        query_text = self._remove_traling_comma(query_text)
        query_text = self._unformat_type_field_name(query_text)
        query_text = re.sub(r'\s+', ' ', query_text).strip()
        return query_text

    def cleanup_query_text(self, query_text: str) -> list[str]:
        """Removes comments and converts text to lines."""
        query_lines = query_text.split('\n')
        result: list[str] = []
        for line in query_lines:
            if re.match('^(#|--|//)', line):
                continue
            cleaned_query_line = re.sub(';$', '',
                                        re.sub('(--|//).*$', '', line).strip())
            result.append(cleaned_query_line)
        return result

    def extract_resource_from_query(self, query: str) -> str:
        return str(
            re.findall(r'FROM\s+([\w.]+)', query,
                       flags=re.IGNORECASE)[0]).strip()

    def extract_query_lines(self, query_text: str) -> list[str]:
        selected_fields = re.sub(
            r'\bSELECT\b|FROM .*', '', query_text,
            flags=re.IGNORECASE).split(',')
        return [field.strip() for field in selected_fields if field != ' ']

    def extract_from_statement(self, query_text: str) -> str:
        return re.search(' FROM .+', query_text, re.IGNORECASE).group(0)

    def _extract_line_elements(self, query_line: str,
                               macros: dict) -> ExtractedLineElements:
        fields: list[str] = []
        virtual_column = None
        field_raw, *alias = re.split(' [Aa][Ss] ', query_line)
        field_raw = field_raw.replace(r'\s+', '').strip()
        processed_field = self._process_field(field_raw)
        virtual_field = processed_field.field
        try:
            operator.attrgetter(virtual_field)(self.base_client.google_ads_row)
        except AttributeError:
            if virtual_field.isdigit():
                virtual_field = int(virtual_field)
            else:
                try:
                    virtual_field = float(virtual_field)
                except ValueError:
                    pass

            operators = ('/', r'\*', r'\+', ' - ')
            if len(expressions := re.split('|'.join(operators), field_raw)) > 1:
                virtual_column_fields = []
                substitute_expression = virtual_field
                for element in expressions:
                    element = element.strip()
                    try:
                        operator.attrgetter(element)(
                            self.base_client.google_ads_row)
                        virtual_column_fields.append(element)
                        substitute_expression = substitute_expression.replace(
                            element, f'{{{element}}}')
                    except AttributeError:
                        pass
                virtual_column = VirtualColumn(
                    type='expression',
                    value=virtual_field.format(
                        **macros) if macros else virtual_field,
                    fields=virtual_column_fields,
                    substitute_expression=substitute_expression.replace(
                        '.', '_'))
            else:
                if not isinstance(virtual_field, (int, float)):
                    if not self._not_a_quoted_string(virtual_field):
                        raise exceptions.GaarfFieldException(
                            f"Incorrect field '{virtual_field}' in the query '{self.text}'."
                        )
                    virtual_field = virtual_field.replace("'",
                                                          '').replace('"', '')
                    virtual_field = virtual_field.format(
                        **macros) if macros else virtual_field
                virtual_column = VirtualColumn(
                    type='built-in', value=virtual_field)
        if not virtual_column and field_raw:
            fields = [field_raw]
        else:
            fields = None
        return ExtractedLineElements(
            fields=fields,
            alias=alias[0] if alias else None,
            virtual_column=virtual_column)

    def _process_field(self, raw_field: str) -> ProcessedField:
        """Process field to extract possible customizers.

        Args:
            field: Unformatted field string value.
        Returns:
            ProcessedField that contains formatted field with customizers.
        """
        if len(resources := self.extract_resource_element(raw_field)) > 1:
            field_name, resource_index = resources
            return ProcessedField(
                field=field_name,
                customizer_type='resource_index',
                customizer_value=int(resource_index))

        if len(nested_fields := self.extract_nested_resource(raw_field)) > 1:
            field_name, nested_field = nested_fields
            return ProcessedField(
                field=field_name,
                customizer_type='nested_field',
                customizer_value=nested_field)
        if len(pointers := self.extract_pointer(raw_field)) > 1:
            field_name, pointer = pointers
            return ProcessedField(
                field=field_name,
                customizer_type='pointer',
                customizer_value=pointer)
        return ProcessedField(field=raw_field)

    def extract_resource_element(self, line_elements: str) -> list[str]:
        return re.split('~', line_elements)

    def extract_pointer(self, line_elements: str) -> list[str]:
        return re.split('->', line_elements)

    def extract_nested_resource(self, line_elements: str) -> list[str]:
        return re.split(':', line_elements)

    def format_type_field_name(self, field_name: str) -> str:
        return re.sub(r'\.type', '.type_', field_name)

    def normalize_column_name(self, column_name: str) -> str:
        return re.sub(r'\.', '_', column_name)

    def _remove_traling_comma(self, query: str) -> str:
        return re.sub(r',\s+from', ' FROM', query, re.IGNORECASE)

    def _unformat_type_field_name(self, query: str) -> str:
        return re.sub(r'\.type_', '.type', query)

    def _not_a_quoted_string(self, field_name: str) -> bool:
        if ((field_name.startswith("'") and field_name.endswith("'")) or
            (field_name.startswith('"') and field_name.endswith('"'))):
            return True
        return False



from __future__ import annotations

import datetime

import pytest
from gaarf import exceptions
from gaarf import query_editor


@pytest.fixture
def query():
    query = """
-- Comment
# Comment
// Comment

SELECT
    1 AS constant,
    '2023-01-01' AS date,
    '{current_date}' AS current_date,
    metrics.clicks / metrics.impressions AS ctr,
    customer.id, --customer_id
    campaign.bidding_strategy_type AS campaign_type, campaign.id:nested AS campaign,
    ad_group.id~1 AS ad_group,
    ad_group_ad.ad.id->asset AS ad,
    metrics.cost_micros * 1e6 AS cost,
    {% if selective == "true" %}
        campaign.selective_optimization AS selective_optimization,
    {% endif %}
from ad_group_ad;
"""
    return query


@pytest.fixture
def builtin_query():
    return 'SELECT * FROM builtin.ocid_mapping'


def test_builtin_query_returns_valid_specification(builtin_query):
    spec = query_editor.QuerySpecification(title='/tmp/ocid_mapping.sql',
                                           text=builtin_query,
                                           args=None).generate()
    assert spec.is_builtin_query
    assert spec.query_title == 'ocid_mapping'
    assert spec.resource_name == 'builtin.ocid_mapping'
    assert not spec.column_names


@pytest.fixture
def query_specification(query):
    return query_editor.QuerySpecification(title='sample_query',
                                           text=query,
                                           args=None)


@pytest.fixture
def query_specification_template(query):
    return query_editor.QuerySpecification(
        title='templated_query',
        text=query,
        args={'template': {
            'selective': 'true'
        }})


@pytest.fixture
def sample_query(query_specification):
    return query_specification.generate()


@pytest.fixture
def templated_query(query_specification_template):
    return query_specification_template.generate()


class TestRegularQuery:

    def test_correct_title(self, sample_query):
        assert sample_query.query_title == 'sample_query'

    def test_extract_correct_fields(self, sample_query):
        assert sample_query.fields == [
            'customer.id', 'campaign.bidding_strategy_type', 'campaign.id',
            'ad_group.id', 'ad_group_ad.ad.id'
        ]

    def test_extract_correct_aliases(self, sample_query):
        assert sample_query.column_names == [
            'constant', 'date', 'current_date', 'ctr', 'customer_id',
            'campaign_type', 'campaign', 'ad_group', 'ad', 'cost'
        ]

    def test_extract_correct_text(self, sample_query):
        assert sample_query.query_text.lower(
        ) == 'select customer.id, campaign.bidding_strategy_type, campaign.id, ad_group.id, ad_group_ad.ad.id, metrics.clicks, metrics.impressions, metrics.cost_micros from ad_group_ad'

    def test_extract_custom_callers(self, sample_query):
        assert sample_query.customizers == {
            'campaign': {
                'type': 'nested_field',
                'value': 'nested'
            },
            'ad_group': {
                'type': 'resource_index',
                'value': 1
            },
            'ad': {
                'type': 'pointer',
                'value': 'asset'
            }
        }

    def test_extract_query_lines(self, query_specification, sample_query):
        extracted_lines = query_specification.extract_query_lines(
            sample_query.query_text)
        assert extracted_lines == [
            'customer.id', 'campaign.bidding_strategy_type', 'campaign.id',
            'ad_group.id', 'ad_group_ad.ad.id', 'metrics.clicks',
            'metrics.impressions', 'metrics.cost_micros'
        ]

    def test_extract_correct_resource(self, sample_query):
        assert sample_query.resource_name == 'ad_group_ad'

    def test_is_constant_resource(self, sample_query):
        assert sample_query.is_constant_resource == False

    def test_has_virtual_columns(self, sample_query):
        assert sample_query.virtual_columns == {
            'constant':
            query_editor.VirtualColumn(type='built-in', value=1),
            'date':
            query_editor.VirtualColumn(type='built-in', value='2023-01-01'),
            'current_date':
            query_editor.VirtualColumn(
                type='built-in',
                value=datetime.date.today().strftime('%Y-%m-%d')),
            'ctr':
            query_editor.VirtualColumn(
                type='expression',
                value='metrics.clicks / metrics.impressions',
                fields=['metrics.clicks', 'metrics.impressions'],
                substitute_expression='{metrics_clicks} / {metrics_impressions}'
            ),
            'cost':
            query_editor.VirtualColumn(
                type='expression',
                value='metrics.cost_micros * 1e6',
                fields=['metrics.cost_micros'],
                substitute_expression='{metrics_cost_micros} * 1e6')
        }

    def test_incorrect_resource_raises_value_error(self):
        query = 'SELECT metrics.clicks FROM ad_groups'
        spec = query_editor.QuerySpecification(title='sample_query',
                                               text=query,
                                               args=None)
        with pytest.raises(exceptions.GaarfResourceException):
            spec.generate()

    def test_incorrect_specification_raises_macro_error(self):
        query = "SELECT '${custom_field}', ad_group.id FROM ad_group"
        spec = query_editor.QuerySpecification(title='sample_query',
                                               text=query,
                                               args=None)
        with pytest.raises(exceptions.GaarfMacroException):
            spec.generate()

    def test_incorrect_specification_raises_virtual_column_error(self):
        query = 'SELECT 1, ad_group.id AS ad_group_id FROM ad_group'
        spec = query_editor.QuerySpecification(title='sample_query',
                                               text=query,
                                               args=None)
        with pytest.raises(exceptions.GaarfVirtualColumnException):
            spec.generate()

    def test_incorrect_field_raises_value_error(self):
        query = 'SELECT metric.impressions, ad_group.id FROM ad_group'
        spec = query_editor.QuerySpecification(title='sample_query',
                                               text=query,
                                               args=None)
        with pytest.raises(exceptions.GaarfFieldException):
            spec.generate()


class TestTemplatedQuery:

    def test_extract_correct_fields(self, templated_query):
        assert templated_query.fields == [
            'customer.id', 'campaign.bidding_strategy_type', 'campaign.id',
            'ad_group.id', 'ad_group_ad.ad.id',
            'campaign.selective_optimization'
        ]

    def test_extract_correct_aliases(self, templated_query):
        assert templated_query.column_names == [
            'constant', 'date', 'current_date', 'ctr', 'customer_id',
            'campaign_type', 'campaign', 'ad_group', 'ad', 'cost',
            'selective_optimization'
        ]

    def test_extract_correct_text(self, templated_query):
        assert templated_query.query_text.lower(
        ) == 'select customer.id, campaign.bidding_strategy_type, campaign.id, ad_group.id, ad_group_ad.ad.id, campaign.selective_optimization, metrics.clicks, metrics.impressions, metrics.cost_micros from ad_group_ad'

    def test_extract_query_lines(self, query_specification, templated_query):
        extracted_lines = query_specification.extract_query_lines(
            templated_query.query_text)
        assert extracted_lines == [
            'customer.id', 'campaign.bidding_strategy_type', 'campaign.id',
            'ad_group.id', 'ad_group_ad.ad.id',
            'campaign.selective_optimization', 'metrics.clicks',
            'metrics.impressions', 'metrics.cost_micros'
        ]



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Module for writing data with console."""
from __future__ import annotations

import rich
from gaarf.io.writers.abs_writer import AbsWriter
from gaarf.report import GaarfReport
from rich.console import Console
from rich.table import Table


class ConsoleWriter(AbsWriter):

    def __init__(self, page_size: int = 10, **kwargs):
        super().__init__(**kwargs)
        self.page_size = int(page_size)

    def write(self, report: GaarfReport, destination: str) -> None:
        report = self.format_for_write(report)
        console = Console()
        table = Table(
            title=f"showing results for query <{destination.split('/')[-1]}>",
            caption=
            f'showing rows 1-{min(self.page_size, len(report))} out of total {len(report)}',
            box=rich.box.MARKDOWN)
        for header in report.column_names:
            table.add_column(header)
        for i, row in enumerate(report):
            if i < self.page_size:
                table.add_row(*[str(field) for field in row])
        console.print(table)



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Module for defining exceptions."""
from __future__ import annotations


class GaarfException(Exception):
    """Base exception."""


class GaarfCliException(GaarfException):
    """Base exception for CLI tools."""


class GaarfMissingQueryException(GaarfCliException):
    """Specifies missing queries to CLI tool."""


class GaarfMissingAccountException(GaarfCliException):
    """Specifies missing account parameter for CLI tool."""


class GaarfQueryException(GaarfException):
    """Base exception for Gaarf queries."""


class GaarfParserException(GaarfException):
    """Base exception for Gaarf parsers."""


class GaarfCustomizerException(GaarfParserException):
    """Specifies incorrect customizer."""


class GaarfVirtualColumnException(GaarfParserException):
    """Specifies incorrect virtual column type."""


class GaarfFieldException(GaarfQueryException):
    """Specifies incorrect Google Ads API field."""


class GaarfMacroException(GaarfQueryException):
    """Specifies incorrect macro in Gaarf query."""


class GaarfResourceException(GaarfQueryException):
    """Specifies incorrect resource name in Google Ads API."""


class GaarfReportException(GaarfException):
    """Base exception for Gaarf reports."""


class GaarfBuiltInQueryException(GaarfException):
    """Specifies missing built-in Gaarf query."""


class GaarfExecutorException(GaarfException):
    """Base exception for errors during GaarfQuery execution."""



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Module for defining base class for Gaarf query classes.

Gaarf query classes can inherit from BaseQuery and have a simple way of
being fetched from API.
"""
from __future__ import annotations


class BaseQuery:
    """Base class to inherit all Gaarf queries from.


    Attributes:
        query_text: Contains query text or template.
    """

    query_text = ''

    @property
    def query(self) -> str:
        """Returns expanded query with parameters takes from initialization."""
        return self.query_text.format(**self.__dict__)

    def __str__(self) -> str:
        """Formatted query string representation."""
        return self.query.strip()



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Module for generating simulated GaarfReport based on query text."""
from __future__ import annotations

import dataclasses
import random
import string
from collections.abc import Sequence
from typing import Any

from faker import Faker
from gaarf.api_clients import BaseClient
from gaarf.api_clients import GOOGLE_ADS_API_VERSION
from gaarf.query_editor import QuerySpecification
from gaarf.query_executor import AdsReportFetcher
from gaarf.report import GaarfReport


@dataclasses.dataclass(frozen=True)
class FakeField:
    type: Any
    name: str


@dataclasses.dataclass
class SimulatorSpecification:
    """Specifies meta information for simulated results."""
    api_version: str = GOOGLE_ADS_API_VERSION
    n_rows: int = 1000
    days_ago: str = '-7d'
    string_length: int = 3
    ignored_enums: tuple[str, ...] = ('REMOVED', 'UNKNOWN', 'UNSPECIFIED')
    allowed_enums: dict[str, Sequence[str]] | None = None
    replacements: dict[str, Any] | None = None


def simulate_data(
    query_text: str,
    query_name: str,
    args: dict,
    api_version: str,
    simulator_specification: SimulatorSpecification,
) -> GaarfReport | None:
    """Simulates GaarfReport based for a given query.

    Args:
        query_text: GAQL query text.
        query_name: Name of the query.
        args: Query parameters to be replaced in query_text.
        api_version: Ads API version to simulate against.
        simulator_specification: Meta information for simulated results.

    Returns:
        Report with simulated data. For `built-in` queries returns None.
    """
    query_specification = QuerySpecification(query_text, query_name, args,
                                             api_version).generate()
    if query_specification.is_builtin_query:
        return None
    client = BaseClient(api_version)
    report_fetcher = AdsReportFetcher(client)
    inferred_types = client.infer_types(query_specification.fields)
    report = report_fetcher.fetch(query_specification)
    try:
        iter(report.results_placeholder[0])
        results = report.results_placeholder[0]
    except TypeError:
        results = [report.results_placeholder[0]]
    if not results:
        results = [results]
    row_field_types = [
        FakeField(type(field_value), field_name)
        for field_value, field_name in zip(results, query_specification.fields)
    ]
    values = _simulate_values(simulator_specification, query_specification,
                              row_field_types, inferred_types)
    return GaarfReport(values, query_specification.column_names)


def _simulate_values(simulator_specification: SimulatorSpecification,
                     query_specification: QuerySpecification,
                     row_field_types: list, inferred_types: list) -> list[list]:
    """Simulates values for a given query based on API version.

    Args:
        simulator_specification: Meta information for simulated results.
        query_specification: Specification of the query.
        row_field_types: Mapping between type of each row element and its name.
        inferred_types: Allowed types for each field of the query.

    Returns:
        Nested list with simulated values.
    """
    simulation_helper = Faker()
    values = []
    for _ in range(simulator_specification.n_rows):
        row = []
        for i, field in enumerate(row_field_types):
            if len(enums := inferred_types[i].values) > 1:
                if allowed_enums := simulator_specification.allowed_enums:
                    if selected_enums := allowed_enums.get(field.name):
                        value = random.choice(selected_enums)
                    else:
                        value = random.choice(list(enums))
                else:
                    value = random.choice([
                        value for value in inferred_types[i].values
                        if value not in simulator_specification.ignored_enums
                    ])
            elif (simulator_specification.replacements and
                  field.name in simulator_specification.replacements):
                value = random.choice(
                    simulator_specification.replacements[field.name])
            else:
                value = _generate_random_value(
                    field, simulator_specification, simulation_helper,
                    query_specification.column_names[i])
            row.append(value)
        values.append(row)
    return values


def _generate_random_value(field: FakeField,
                           simulator_specification: SimulatorSpecification,
                           simulation_helper: Faker,
                           column_name: str) -> bool | str | float:
    """Generates random values based on field type and name.

    Args:
        field: A single element of a row.
        simulator_specification: Meta information for simulated results.
        simulation_helper: Class to perform simulation operations.
        column_name: Name of the column that needs to be simulated.

    Returns:
        Simulated value.
    """
    if field.type == str:
        if 'date' in field.name:
            return simulation_helper.date_between(
                simulator_specification.days_ago).strftime('%Y-%m-%d')
        if 'url' in field.name:
            return 'example.com'
        if 'video.id' in field.name or 'video_id' in field.name:
            return '4WXs3sKu41I'
        if 'id' in column_name:
            return str(random.randint(1000000, 1000010))
        return _generate_random_string(simulator_specification.string_length)
    if field.type == int:
        if '.id' in field.name:
            return random.randint(1000000, 1000010)
        if 'micros' in field.name:
            return int(random.randint(0, 1000) * 1e6)
        return random.randint(0, 1000)
    if field.type == float:
        return float(random.randint(0, 1000))
    if field.type == bool:
        return bool(random.getrandbits(1))
    return ''


def _generate_random_string(length: int) -> str:
    'Helper for simulating random string of a given length.' ''
    characters = string.ascii_lowercase
    return ''.join(random.choice(characters) for _ in range(length))
